from jivago.lang.stream import Stream


class UrlEncodedQueryParser(object):

    def parse_urlencoded_query(self, query_string: str) -> dict:
        pairs = query_string.split("&")
        return Stream(pairs).filter(lambda x: x != "").map(lambda pair: pair.split("=")).toDict()



from typing import List, Type

from jivago.config.production_jivago_context import ProductionJivagoContext
from jivago.lang.annotations import Override
from jivago.lang.registry import Registry
from jivago.templating.template_filter import TemplateFilter
from jivago.wsgi.filters.body_serialization_filter import BodySerializationFilter
from jivago.wsgi.filters.exception.application_exception_filter import ApplicationExceptionFilter
from jivago.wsgi.filters.exception.debug_exception_filter import DebugExceptionFilter
from jivago.wsgi.filters.jivago_banner_filter import JivagoBannerFilter
from jivago.wsgi.filters.exception.unknown_exception_filter import UnknownExceptionFilter
from jivago.wsgi.filters.filter import Filter
from jivago.wsgi.request.http_form_deserialization_filter import HttpFormDeserializationFilter
from jivago.wsgi.request.json_serialization_filter import JsonSerializationFilter
from jivago.wsgi.request.no_cors_filter import NoCorsFilter


class DebugJivagoContext(ProductionJivagoContext):

    def __init__(self, root_package: "Module", registry: Registry):
        super().__init__(root_package, registry)

    @Override
    def get_filters(self, path: str) -> List[Type[Filter]]:
        return [JivagoBannerFilter, NoCorsFilter, UnknownExceptionFilter, DebugExceptionFilter, TemplateFilter, JsonSerializationFilter,
                HttpFormDeserializationFilter, BodySerializationFilter, ApplicationExceptionFilter]



from jivago.lang.annotations import Inject
from jivago.wsgi.dto_serialization_handler import DtoSerializationHandler
from jivago.wsgi.filters.filter import Filter
from jivago.wsgi.filters.filter_chain import FilterChain
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response


class BodySerializationFilter(Filter):

    @Inject
    def __init__(self, dto_serialization_handler: DtoSerializationHandler):
        self.dto_serialization_handler = dto_serialization_handler

    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        chain.doFilter(request, response)

        if self.dto_serialization_handler.is_serializable(response.body):
            response.body = self.dto_serialization_handler.serialize(response.body)



from jivago.lang.annotations import Override, Inject
from jivago.wsgi.filters.filter import Filter
from jivago.wsgi.filters.filter_chain import FilterChain
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response
from jivago.wsgi.request.url_encoded_query_parser import UrlEncodedQueryParser


class HttpFormDeserializationFilter(Filter):

    @Inject
    def __init__(self, query_parser: UrlEncodedQueryParser):
        self.queryParser = query_parser

    @Override
    def doFilter(self, request: Request, response: Response, chain: FilterChain):
        if request.headers['Content-Type'] == 'application/x-www-form-urlencoded':
            request.body = self.queryParser.parse_urlencoded_query(request.body.decode("utf-8"))

        chain.doFilter(request, response)



import urllib.parse
from typing import TypingMeta, _Union

from jivago.inject.service_locator import ServiceLocator
from jivago.wsgi.dto_serialization_handler import DtoSerializationHandler
from jivago.wsgi.incorrect_resource_parameters_exception import IncorrectResourceParametersException
from jivago.wsgi.methods import to_method
from jivago.wsgi.missing_route_invocation_argument import MissingRouteInvocationArgument
from jivago.wsgi.request.request import Request
from jivago.wsgi.request.response import Response
from jivago.wsgi.request.url_encoded_query_parser import UrlEncodedQueryParser
from jivago.wsgi.route_registration import RouteRegistration
from jivago.wsgi.routing_table import RoutingTable

ALLOWED_URL_PARAMETER_TYPES = (str, int, float)


class ResourceInvocator(object):

    def __init__(self, service_locator: ServiceLocator, routing_table: RoutingTable,
                 dto_serialization_handler: DtoSerializationHandler, query_parser: UrlEncodedQueryParser):
        self.dto_serialization_handler = dto_serialization_handler
        self.routing_table = routing_table
        self.service_locator = service_locator
        self.query_parser = query_parser

    def invoke(self, request: Request) -> Response:
        method = to_method(request.method)
        for route_registration in self.routing_table.get_route_registration(method, request.path):
            resource = self.service_locator.get(route_registration.resourceClass)
            try:
                parameters = self.format_parameters(request, route_registration)
                function_return = route_registration.routeFunction(resource, *parameters)
            except (MissingRouteInvocationArgument, AttributeError):
                continue

            if isinstance(function_return, Response):
                return function_return
            return Response(200, {}, function_return)
        raise IncorrectResourceParametersException()

    def format_parameters(self, request: Request, route_registration: RouteRegistration) -> list:
        parameter_declaration = route_registration.routeFunction.__annotations__.items()
        path_parameters = route_registration.parse_path_parameters(request.path)
        query_parameters = self.query_parser.parse_urlencoded_query(request.queryString)

        parameters = []
        for name, clazz in parameter_declaration:
            if name == 'return':  # This is the output type annotation
                break
            if isinstance(clazz, _Union) and type(None) in clazz.__args__:
                parameters.append(self.__get_single_parameter(name, clazz.__args__[0], request, path_parameters, query_parameters, nullable=True))
            else:
                parameters.append(self.__get_single_parameter(name, clazz, request, path_parameters, query_parameters, nullable=False))

        return parameters

    def __get_single_parameter(self, parameter_name: str, parameter_type: type, request: Request, path_parameters: dict, query_parameters: dict,
                               nullable: bool) -> object:
        if parameter_type == Request:
            return request
        elif parameter_type == dict:
            return request.body
        elif parameter_type in ALLOWED_URL_PARAMETER_TYPES:
            if parameter_name in path_parameters:
                return parameter_type(self._url_parameter_unescape(path_parameters[parameter_name]))
            elif parameter_name in query_parameters:
                return parameter_type(self._url_parameter_unescape(query_parameters[parameter_name]))
        elif self.dto_serialization_handler.is_deserializable_into(parameter_type):
            return self.dto_serialization_handler.deserialize(request.body, parameter_type)
        if nullable:
            return None
        raise MissingRouteInvocationArgument(parameter_name, parameter_type)

    def _url_parameter_unescape(self, escaped):
        return urllib.parse.unquote(escaped)



import os
from typing import List, Type

from jivago.config.abstract_context import AbstractContext
from jivago.config.exception_mapper_binder import ExceptionMapperBinder
from jivago.config.startup_hooks import Init, PreInit, PostInit
from jivago.inject.annoted_class_binder import AnnotatedClassBinder
from jivago.inject.provider_binder import ProviderBinder
from jivago.inject.scope_cache import ScopeCache
from jivago.lang.annotations import Override, BackgroundWorker
from jivago.lang.registry import Singleton, Component, Registry
from jivago.lang.stream import Stream
from jivago.scheduling.annotations import Scheduled
from jivago.scheduling.task_scheduler import TaskScheduler
from jivago.templating.template_filter import TemplateFilter
from jivago.templating.view_template_repository import ViewTemplateRepository
from jivago.wsgi.annotations import Resource
from jivago.wsgi.dto_serialization_handler import DtoSerializationHandler
from jivago.wsgi.filters.body_serialization_filter import BodySerializationFilter
from jivago.wsgi.filters.exception.application_exception_filter import ApplicationExceptionFilter
from jivago.wsgi.filters.exception.unknown_exception_filter import UnknownExceptionFilter
from jivago.wsgi.filters.filter import Filter
from jivago.wsgi.http_status_code_resolver import HttpStatusCodeResolver
from jivago.wsgi.partial_content_handler import PartialContentHandler
from jivago.wsgi.request.http_form_deserialization_filter import HttpFormDeserializationFilter
from jivago.wsgi.request.json_serialization_filter import JsonSerializationFilter
from jivago.wsgi.request.url_encoded_query_parser import UrlEncodedQueryParser


class ProductionJivagoContext(AbstractContext):

    def __init__(self, root_package: "Module", registry: Registry):
        self.root_package_name = root_package.__name__ if root_package else ''
        self.root_package = root_package
        self.registry = registry
        super().__init__()

    @Override
    def configure_service_locator(self):
        AnnotatedClassBinder(self.root_package_name, self.registry, Component).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, Resource).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, BackgroundWorker).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, Scheduled).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, Init).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, PreInit).bind(self.serviceLocator)
        AnnotatedClassBinder(self.root_package_name, self.registry, PostInit).bind(self.serviceLocator)

        ProviderBinder(self.root_package_name, self.registry).bind(self.serviceLocator)
        for scope in self.scopes():
            scoped_classes = Stream(self.registry.get_annotated_in_package(scope, self.root_package_name)).map(lambda registration: registration.registered).toList()
            cache = ScopeCache(scope, scoped_classes)
            self.serviceLocator.register_scope(cache)

        Stream(self.get_filters("")).forEach(lambda f: self.serviceLocator.bind(f, f))

        # TODO better way to handle Jivago Dependencies
        self.serviceLocator.bind(TaskScheduler, TaskScheduler(self.serviceLocator))
        self.serviceLocator.bind(DtoSerializationHandler,
                                 DtoSerializationHandler(Registry(), self.root_package_name))
        self.serviceLocator.bind(ViewTemplateRepository, ViewTemplateRepository(self.get_views_folder_path()))
        self.serviceLocator.bind(UrlEncodedQueryParser, UrlEncodedQueryParser)
        self.serviceLocator.bind(BodySerializationFilter, BodySerializationFilter)
        self.serviceLocator.bind(PartialContentHandler, PartialContentHandler)
        self.serviceLocator.bind(HttpStatusCodeResolver, HttpStatusCodeResolver)

        ExceptionMapperBinder().bind(self.serviceLocator)

    def scopes(self) -> List[type]:
        return [Singleton, BackgroundWorker]

    @Override
    def get_filters(self, path: str) -> List[Type[Filter]]:
        return [UnknownExceptionFilter, TemplateFilter, JsonSerializationFilter, HttpFormDeserializationFilter, BodySerializationFilter,
                ApplicationExceptionFilter]

    @Override
    def get_views_folder_path(self) -> str:
        return os.path.join(os.path.dirname(self.root_package.__file__), "views") if self.root_package else ''

    @Override
    def get_config_file_locations(self) -> List[str]:
        return ["application.yml", "application.json", "properties.yml", "properties.json"]
